# Copyright (C) 2007-2012, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from breezy import controldir, errors, tests, workingtree
from breezy.tests.script import TestCaseWithTransportAndScript


class TestReconfigure(TestCaseWithTransportAndScript):
    def test_no_type(self):
        self.make_branch("branch")
        self.run_bzr_error(["No target configuration specified"], "reconfigure branch")

    def test_branch_to_tree(self):
        self.make_branch("branch")
        self.run_bzr("reconfigure --tree branch")
        workingtree.WorkingTree.open("branch")

    def test_tree_to_branch(self):
        self.make_branch_and_tree("tree")
        self.run_bzr("reconfigure --branch tree")
        self.assertRaises(errors.NoWorkingTree, workingtree.WorkingTree.open, "tree")

    def test_branch_to_specified_checkout(self):
        self.make_branch("branch")
        self.make_branch("parent")
        self.run_bzr("reconfigure branch --checkout --bind-to parent")

    def test_force(self):
        tree = self.make_branch_and_tree("tree")
        self.build_tree(["tree/file"])
        tree.add("file")
        self.run_bzr_error(
            ['Working tree ".*" has uncommitted changes'], "reconfigure --branch tree"
        )
        self.run_bzr("reconfigure --force --branch tree")

    def test_lightweight_checkout_to_checkout(self):
        branch = self.make_branch("branch")
        branch.create_checkout("checkout", lightweight=True)
        self.run_bzr("reconfigure --checkout checkout")

    def test_lightweight_checkout_to_tree(self):
        branch = self.make_branch("branch")
        branch.create_checkout("checkout", lightweight=True)
        self.run_bzr("reconfigure --tree checkout")

    def test_no_args(self):
        self.make_branch("branch")
        self.run_bzr_error(
            ["No target configuration specified"], "reconfigure", working_dir="branch"
        )

    def test_checkout_to_lightweight_checkout(self):
        branch = self.make_branch("branch")
        branch.create_checkout("checkout")
        self.run_bzr("reconfigure --lightweight-checkout checkout")

    def test_standalone_to_use_shared(self):
        self.build_tree(["repo/"])
        tree = self.make_branch_and_tree("repo/tree")
        self.make_repository("repo", shared=True)
        self.run_bzr("reconfigure --use-shared", working_dir="repo/tree")
        tree = workingtree.WorkingTree.open("repo/tree")
        self.assertNotEqual(
            tree.controldir.root_transport.base,
            tree.branch.repository.controldir.root_transport.base,
        )

    def test_use_shared_to_standalone(self):
        self.make_repository("repo", shared=True)
        branch = controldir.ControlDir.create_branch_convenience("repo/tree")
        self.assertNotEqual(
            branch.controldir.root_transport.base,
            branch.repository.controldir.root_transport.base,
        )
        self.run_bzr("reconfigure --standalone", working_dir="repo/tree")
        tree = workingtree.WorkingTree.open("repo/tree")
        self.assertEqual(
            tree.controldir.root_transport.base,
            tree.branch.repository.controldir.root_transport.base,
        )

    def test_make_with_trees(self):
        repo = self.make_repository("repo", shared=True)
        repo.set_make_working_trees(False)
        self.run_bzr("reconfigure --with-trees", working_dir="repo")
        self.assertIs(True, repo.make_working_trees())

    def test_make_with_trees_already_trees(self):
        repo = self.make_repository("repo", shared=True)
        repo.set_make_working_trees(True)
        self.run_bzr_error(
            [" already creates working trees"], "reconfigure --with-trees repo"
        )

    def test_make_without_trees(self):
        repo = self.make_repository("repo", shared=True)
        repo.set_make_working_trees(True)
        self.run_bzr("reconfigure --with-no-trees", working_dir="repo")
        self.assertIs(False, repo.make_working_trees())

    def test_make_without_trees_already_no_trees(self):
        repo = self.make_repository("repo", shared=True)
        repo.set_make_working_trees(False)
        self.run_bzr_error(
            [" already doesn't create working trees"],
            "reconfigure --with-no-trees repo",
        )

    def test_make_with_trees_nonshared_repo(self):
        self.make_branch("branch")
        self.run_bzr_error(
            ["Requested reconfiguration of '.*' is not supported"],
            "reconfigure --with-trees branch",
        )

    def test_make_without_trees_leaves_tree_alone(self):
        self.make_repository("repo", shared=True)
        controldir.ControlDir.create_branch_convenience("repo/branch")
        tree = workingtree.WorkingTree.open("repo/branch")
        self.build_tree(["repo/branch/foo"])
        tree.add("foo")
        self.run_bzr("reconfigure --with-no-trees --force", working_dir="repo/branch")
        self.assertPathExists("repo/branch/foo")
        tree = workingtree.WorkingTree.open("repo/branch")

    def test_shared_format_to_standalone(self, format=None):
        self.make_repository("repo", shared=True, format=format)
        branch = controldir.ControlDir.create_branch_convenience("repo/tree")
        self.assertNotEqual(
            branch.controldir.root_transport.base,
            branch.repository.controldir.root_transport.base,
        )
        tree = workingtree.WorkingTree.open("repo/tree")
        self.build_tree_contents([("repo/tree/file", b"foo\n")])
        tree.add(["file"])
        tree.commit("added file")
        self.run_bzr("reconfigure --standalone", working_dir="repo/tree")
        tree = workingtree.WorkingTree.open("repo/tree")
        self.build_tree_contents([("repo/tree/file", b"bar\n")])
        self.check_file_contents("repo/tree/file", b"bar\n")
        self.run_bzr("revert", working_dir="repo/tree")
        self.check_file_contents("repo/tree/file", b"foo\n")
        self.assertEqual(
            tree.controldir.root_transport.base,
            tree.branch.repository.controldir.root_transport.base,
        )

    def test_shared_knit_to_standalone(self):
        self.test_shared_format_to_standalone("knit")

    def test_shared_pack092_to_standalone(self):
        self.test_shared_format_to_standalone("pack-0.92")

    def test_shared_rich_root_pack_to_standalone(self):
        self.test_shared_format_to_standalone("rich-root-pack")

    def test_lightweight_format_checkout_to_tree(self, format=None):
        branch = self.make_branch("branch", format=format)
        branch.create_checkout("checkout", lightweight=True)
        tree = workingtree.WorkingTree.open("checkout")
        self.build_tree_contents([("checkout/file", b"foo\n")])
        tree.add(["file"])
        tree.commit("added file")
        self.run_bzr("reconfigure --tree", working_dir="checkout")
        tree = workingtree.WorkingTree.open("checkout")
        self.build_tree_contents([("checkout/file", b"bar\n")])
        self.check_file_contents("checkout/file", b"bar\n")
        self.run_bzr("revert", working_dir="checkout")
        self.check_file_contents("checkout/file", b"foo\n")

    def test_lightweight_knit_checkout_to_tree(self):
        self.test_lightweight_format_checkout_to_tree("knit")

    def test_lightweight_pack092_checkout_to_tree(self):
        self.test_lightweight_format_checkout_to_tree("pack-0.92")

    def test_lightweight_rich_root_pack_checkout_to_tree(self):
        self.test_lightweight_format_checkout_to_tree("rich-root-pack")

    def test_branch_and_use_shared(self):
        self.run_script("""\
$ brz init -q branch
$ echo foo > branch/foo
$ brz add -q branch/foo
$ brz commit -q -m msg branch
$ brz init-shared-repo -q .
$ brz reconfigure --branch --use-shared branch
$ brz info branch
Repository branch (format: ...)
Location:
  shared repository: .
  repository branch: branch
""")

    def test_use_shared_and_branch(self):
        self.run_script("""\
$ brz init -q branch
$ echo foo > branch/foo
$ brz add -q branch/foo
$ brz commit -q -m msg branch
$ brz init-shared-repo -q .
$ brz reconfigure --use-shared --branch branch
$ brz info branch
Repository branch (format: ...)
Location:
  shared repository: .
  repository branch: branch
""")


class TestReconfigureStacking(tests.TestCaseWithTransport):
    def test_reconfigure_stacking(self):
        """Test stacking reconfiguration.

        Test a fairly realistic scenario for stacking:

         * make a branch with some history
         * branch it
         * make the second branch stacked on the first
         * commit in the second
         * then make the second unstacked, so it has to fill in history from
           the original fallback lying underneath its original content

        See discussion in <https://bugs.launchpad.net/bzr/+bug/391411>
        """
        # there are also per_branch tests that exercise remote operation etc
        tree_1 = self.make_branch_and_tree("b1", format="2a")
        self.build_tree(["b1/foo"])
        tree_1.add(["foo"])
        tree_1.commit("add foo")
        # now branch and commit again
        bzrdir_2 = tree_1.controldir.sprout("b2")
        tree_2 = bzrdir_2.open_workingtree()
        branch_2 = tree_2.branch
        # now reconfigure to be stacked
        out, err = self.run_bzr("reconfigure --stacked-on b1 b2")
        self.assertContainsRe(out, "^.*/b2/ is now stacked on ../b1\n$")
        self.assertEqual("", err)
        # can also give the absolute URL of the branch, and it gets stored
        # as a relative path if possible
        out, err = self.run_bzr(
            "reconfigure --stacked-on {} b2".format(self.get_url("b1"))
        )
        self.assertContainsRe(out, "^.*/b2/ is now stacked on ../b1\n$")
        self.assertEqual("", err)
        # Refresh the branch as 'reconfigure' modified it
        branch_2 = branch_2.controldir.open_branch()
        # It should be given a relative URL to the destination, if possible,
        # because that's most likely to work across different transports
        self.assertEqual("../b1", branch_2.get_stacked_on_url())
        # commit, and it should be stored into b2's repo
        self.build_tree_contents([("foo", b"new foo")])
        tree_2.commit("update foo")
        # Now turn it off again
        out, err = self.run_bzr("reconfigure --unstacked b2")
        self.assertContainsRe(out, "^.*/b2/ is now not stacked\n$")
        self.assertEqual("", err)
        # Refresh the branch as 'reconfigure' modified it
        branch_2 = branch_2.controldir.open_branch()
        self.assertRaises(errors.NotStacked, branch_2.get_stacked_on_url)

    # XXX: Needs a test for reconfiguring stacking and shape at the same time;
    # no branch at location; stacked-on is not a branch; quiet mode.
    # -- mbp 20090706



# Copyright (C) 2006-2009, 2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for MultiWalker."""

from breezy import multiwalker, revision
from breezy.tests import TestCaseWithTransport


class TestMultiWalker(TestCaseWithTransport):
    def assertStepOne(self, has_more, path, file_id, iterator):
        retval = multiwalker.MultiWalker._step_one(iterator)
        if not has_more:
            self.assertIs(None, path)
            self.assertIs(None, file_id)
            self.assertEqual((False, None, None), retval)
        else:
            self.assertEqual(
                (has_more, path, file_id), (retval[0], retval[1], retval[2].file_id)
            )

    def test__step_one_empty(self):
        tree = self.make_branch_and_tree("empty")
        repo = tree.branch.repository
        empty_tree = repo.revision_tree(revision.NULL_REVISION)

        iterator = empty_tree.iter_entries_by_dir()
        self.assertStepOne(False, None, None, iterator)
        self.assertStepOne(False, None, None, iterator)

    def test__step_one(self):
        tree = self.make_branch_and_tree("tree")
        self.build_tree(["tree/a", "tree/b/", "tree/b/c"])
        tree.add(["a", "b", "b/c"], ids=[b"a-id", b"b-id", b"c-id"])

        iterator = tree.iter_entries_by_dir()
        tree.lock_read()
        self.addCleanup(tree.unlock)

        root_id = tree.path2id("")
        self.assertStepOne(True, "", root_id, iterator)
        self.assertStepOne(True, "a", b"a-id", iterator)
        self.assertStepOne(True, "b", b"b-id", iterator)
        self.assertStepOne(True, "b/c", b"c-id", iterator)
        self.assertStepOne(False, None, None, iterator)
        self.assertStepOne(False, None, None, iterator)

    def assertWalkerNext(
        self, exp_path, exp_file_id, master_has_node, exp_other_paths, iterator
    ):
        """Check what happens when we step the iterator.

        :param path: The path for this entry
        :param file_id: The file_id for this entry
        :param master_has_node: Does the master tree have this entry?
        :param exp_other_paths: A list of other_path values.
        :param iterator: The iterator to step
        """
        path, file_id, master_ie, other_values = next(iterator)
        self.assertEqual(
            (exp_path, exp_file_id), (path, file_id), "Master entry did not match"
        )
        if master_has_node:
            self.assertIsNot(None, master_ie, "master should have an entry")
        else:
            self.assertIs(None, master_ie, "master should not have an entry")
        self.assertEqual(
            len(exp_other_paths), len(other_values), "Wrong number of other entries"
        )
        other_paths = []
        other_file_ids = []
        for path, ie in other_values:
            other_paths.append(path)
            if ie is None:
                other_file_ids.append(None)
            else:
                other_file_ids.append(ie.file_id)

        exp_file_ids = []
        for path in exp_other_paths:
            if path is None:
                exp_file_ids.append(None)
            else:
                exp_file_ids.append(file_id)
        self.assertEqual(exp_other_paths, other_paths, "Other paths incorrect")
        self.assertEqual(exp_file_ids, other_file_ids, "Other file_ids incorrect")

    def lock_and_get_basis_and_root_id(self, tree):
        tree.lock_read()
        self.addCleanup(tree.unlock)
        basis_tree = tree.basis_tree()
        basis_tree.lock_read()
        self.addCleanup(basis_tree.unlock)
        root_id = tree.path2id("")
        return basis_tree, root_id

    def test_simple_stepping(self):
        tree = self.make_branch_and_tree("tree")
        self.build_tree(["tree/a", "tree/b/", "tree/b/c"])
        tree.add(["a", "b", "b/c"], ids=[b"a-id", b"b-id", b"c-id"])

        tree.commit("first", rev_id=b"first-rev-id")

        basis_tree, root_id = self.lock_and_get_basis_and_root_id(tree)

        walker = multiwalker.MultiWalker(tree, [basis_tree])
        iterator = walker.iter_all()
        self.assertWalkerNext("", root_id, True, [""], iterator)
        self.assertWalkerNext("a", b"a-id", True, ["a"], iterator)
        self.assertWalkerNext("b", b"b-id", True, ["b"], iterator)
        self.assertWalkerNext("b/c", b"c-id", True, ["b/c"], iterator)
        self.assertRaises(StopIteration, next, iterator)

    def test_master_has_extra(self):
        tree = self.make_branch_and_tree("tree")
        self.build_tree(["tree/a", "tree/b/", "tree/c", "tree/d"])
        tree.add(["a", "b", "d"], ids=[b"a-id", b"b-id", b"d-id"])

        tree.commit("first", rev_id=b"first-rev-id")

        tree.add(["c"], ids=[b"c-id"])
        basis_tree, root_id = self.lock_and_get_basis_and_root_id(tree)

        walker = multiwalker.MultiWalker(tree, [basis_tree])
        iterator = walker.iter_all()
        self.assertWalkerNext("", root_id, True, [""], iterator)
        self.assertWalkerNext("a", b"a-id", True, ["a"], iterator)
        self.assertWalkerNext("b", b"b-id", True, ["b"], iterator)
        self.assertWalkerNext("c", b"c-id", True, [None], iterator)
        self.assertWalkerNext("d", b"d-id", True, ["d"], iterator)
        self.assertRaises(StopIteration, next, iterator)

    def test_master_renamed_to_earlier(self):
        """The record is still present, it just shows up early."""
        tree = self.make_branch_and_tree("tree")
        self.build_tree(["tree/a", "tree/c", "tree/d"])
        tree.add(["a", "c", "d"], ids=[b"a-id", b"c-id", b"d-id"])
        tree.commit("first", rev_id=b"first-rev-id")
        tree.rename_one("d", "b")

        basis_tree, root_id = self.lock_and_get_basis_and_root_id(tree)

        walker = multiwalker.MultiWalker(tree, [basis_tree])
        iterator = walker.iter_all()
        self.assertWalkerNext("", root_id, True, [""], iterator)
        self.assertWalkerNext("a", b"a-id", True, ["a"], iterator)
        self.assertWalkerNext("b", b"d-id", True, ["d"], iterator)
        self.assertWalkerNext("c", b"c-id", True, ["c"], iterator)
        self.assertRaises(StopIteration, next, iterator)

    def test_master_renamed_to_later(self):
        tree = self.make_branch_and_tree("tree")
        self.build_tree(["tree/a", "tree/b", "tree/d"])
        tree.add(["a", "b", "d"], ids=[b"a-id", b"b-id", b"d-id"])
        tree.commit("first", rev_id=b"first-rev-id")
        tree.rename_one("b", "e")

        basis_tree, root_id = self.lock_and_get_basis_and_root_id(tree)

        walker = multiwalker.MultiWalker(tree, [basis_tree])
        iterator = walker.iter_all()
        self.assertWalkerNext("", root_id, True, [""], iterator)
        self.assertWalkerNext("a", b"a-id", True, ["a"], iterator)
        self.assertWalkerNext("d", b"d-id", True, ["d"], iterator)
        self.assertWalkerNext("e", b"b-id", True, ["b"], iterator)
        self.assertRaises(StopIteration, next, iterator)

    def test_other_extra_in_middle(self):
        tree = self.make_branch_and_tree("tree")
        self.build_tree(["tree/a", "tree/b", "tree/d"])
        tree.add(["a", "b", "d"], ids=[b"a-id", b"b-id", b"d-id"])
        tree.commit("first", rev_id=b"first-rev-id")
        tree.remove(["b"])

        basis_tree, root_id = self.lock_and_get_basis_and_root_id(tree)
        walker = multiwalker.MultiWalker(tree, [basis_tree])
        iterator = walker.iter_all()
        self.assertWalkerNext("", root_id, True, [""], iterator)
        self.assertWalkerNext("a", b"a-id", True, ["a"], iterator)
        self.assertWalkerNext("d", b"d-id", True, ["d"], iterator)
        self.assertWalkerNext("b", b"b-id", False, ["b"], iterator)
        self.assertRaises(StopIteration, next, iterator)

    def test_other_extra_at_end(self):
        tree = self.make_branch_and_tree("tree")
        self.build_tree(["tree/a", "tree/b", "tree/d"])
        tree.add(["a", "b", "d"], ids=[b"a-id", b"b-id", b"d-id"])
        tree.commit("first", rev_id=b"first-rev-id")
        tree.remove(["d"])

        basis_tree, root_id = self.lock_and_get_basis_and_root_id(tree)
        walker = multiwalker.MultiWalker(tree, [basis_tree])
        iterator = walker.iter_all()
        self.assertWalkerNext("", root_id, True, [""], iterator)
        self.assertWalkerNext("a", b"a-id", True, ["a"], iterator)
        self.assertWalkerNext("b", b"b-id", True, ["b"], iterator)
        self.assertWalkerNext("d", b"d-id", False, ["d"], iterator)
        self.assertRaises(StopIteration, next, iterator)

    def test_others_extra_at_end(self):
        tree = self.make_branch_and_tree("tree")
        self.build_tree(["tree/a", "tree/b", "tree/c", "tree/d", "tree/e"])
        tree.add(
            ["a", "b", "c", "d", "e"], ids=[b"a-id", b"b-id", b"c-id", b"d-id", b"e-id"]
        )
        tree.commit("first", rev_id=b"first-rev-id")
        tree.remove(["e"])
        tree.commit("second", rev_id=b"second-rev-id")
        tree.remove(["d"])
        tree.commit("third", rev_id=b"third-rev-id")
        tree.remove(["c"])

        basis_tree, root_id = self.lock_and_get_basis_and_root_id(tree)
        first_tree = tree.branch.repository.revision_tree(b"first-rev-id")
        second_tree = tree.branch.repository.revision_tree(b"second-rev-id")
        walker = multiwalker.MultiWalker(tree, [basis_tree, first_tree, second_tree])
        iterator = walker.iter_all()
        self.assertWalkerNext("", root_id, True, ["", "", ""], iterator)
        self.assertWalkerNext("a", b"a-id", True, ["a", "a", "a"], iterator)
        self.assertWalkerNext("b", b"b-id", True, ["b", "b", "b"], iterator)
        self.assertWalkerNext("c", b"c-id", False, ["c", "c", "c"], iterator)
        self.assertWalkerNext("d", b"d-id", False, [None, "d", "d"], iterator)
        self.assertWalkerNext("e", b"e-id", False, [None, "e", None], iterator)
        self.assertRaises(StopIteration, next, iterator)

    def test_different_file_id_in_others(self):
        tree = self.make_branch_and_tree("tree")
        self.build_tree(["tree/a", "tree/b", "tree/c/"])
        tree.add(["a", "b", "c"], ids=[b"a-id", b"b-id", b"c-id"])
        tree.commit("first", rev_id=b"first-rev-id")

        tree.rename_one("b", "c/d")
        self.build_tree(["tree/b"])
        tree.add(["b"], ids=[b"b2-id"])
        tree.commit("second", rev_id=b"second-rev-id")

        tree.rename_one("a", "c/e")
        self.build_tree(["tree/a"])
        tree.add(["a"], ids=[b"a2-id"])

        basis_tree, root_id = self.lock_and_get_basis_and_root_id(tree)
        first_tree = tree.branch.repository.revision_tree(b"first-rev-id")
        walker = multiwalker.MultiWalker(tree, [basis_tree, first_tree])

        iterator = walker.iter_all()
        self.assertWalkerNext("", root_id, True, ["", ""], iterator)
        self.assertWalkerNext("a", b"a2-id", True, [None, None], iterator)
        self.assertWalkerNext("b", b"b2-id", True, ["b", None], iterator)
        self.assertWalkerNext("c", b"c-id", True, ["c", "c"], iterator)
        self.assertWalkerNext("c/d", b"b-id", True, ["c/d", "b"], iterator)
        self.assertWalkerNext("c/e", b"a-id", True, ["a", "a"], iterator)
        self.assertRaises(StopIteration, next, iterator)

    def assertLtByDirblock(self, lt_val, path1, path2):
        self.assertEqual(
            lt_val, multiwalker.MultiWalker._lt_path_by_dirblock(path1, path2)
        )

    def test__lt_path_by_dirblock(self):
        # We only support Unicode strings at this point
        self.assertRaises(
            TypeError, multiwalker.MultiWalker._lt_path_by_dirblock, b"", b"b"
        )
        self.assertLtByDirblock(False, "", "")
        self.assertLtByDirblock(False, "a", "a")
        self.assertLtByDirblock(False, "a/b", "a/b")
        self.assertLtByDirblock(False, "a/b/c", "a/b/c")
        self.assertLtByDirblock(False, "a-a", "a")
        self.assertLtByDirblock(True, "a-a", "a/a")
        self.assertLtByDirblock(True, "a=a", "a/a")
        self.assertLtByDirblock(False, "a-a/a", "a/a")
        self.assertLtByDirblock(False, "a=a/a", "a/a")
        self.assertLtByDirblock(False, "a-a/a", "a/a/a")
        self.assertLtByDirblock(False, "a=a/a", "a/a/a")
        self.assertLtByDirblock(False, "a-a/a/a", "a/a/a")
        self.assertLtByDirblock(False, "a=a/a/a", "a/a/a")

    def assertPathToKey(self, expected, path):
        self.assertEqual(expected, multiwalker.MultiWalker._path_to_key(path))

    def test__path_to_key(self):
        self.assertPathToKey(([""], ""), "")
        self.assertPathToKey(([""], "a"), "a")
        self.assertPathToKey((["a"], "b"), "a/b")
        self.assertPathToKey((["a", "b"], "c"), "a/b/c")



# Copyright (C) 2008 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Custom module finder for entire package."""

import os
import sys

# At present, this is only used on Windows (see setup.py)
from py2exe import mf310 as modulefinder


class CustomModuleFinder(modulefinder.ModuleFinder):
    """Custom module finder for processing python packages,
    e.g. brz plugins packages.

    :param  path:   list of directories to search for modules;
                    if not specified, python standard library only is used.
    """

    def __init__(self, path=None, debug=0, excludes=None, replace_paths=None):
        if replace_paths is None:
            replace_paths = []
        if excludes is None:
            excludes = []
        if path is None:
            path = [os.path.dirname(os.__file__)]  # only python std lib
        modulefinder.ModuleFinder.__init__(self, path, debug, excludes, replace_paths)

    def load_package_recursive(self, fqname):
        """Recursively process each module in package.

        :param  fqname:   name of the package.
        """
        # Load all the parents
        parent = None
        path = []
        for partname in fqname.split("."):
            parent_path = ".".join(path)
            path.append(partname)
            # import_module works recursively,
            # and some of the dependencies may try
            # to import modules not present on the system.
            # (The actual error is
            # AttributeError: 'NoneType' object has no attribute 'is_package')
            # Ignore errors here and bail out in the collection loop.
            try:
                self.import_module(
                    partname, ".".join(path), self.modules.get(parent_path, None)
                )
            except BaseException:
                pass
        stack = [(fqname, parent_path)]
        while stack:
            (package, parent_path) = stack.pop(0)
            # Here we assume that all parents have already been imported.
            # Abort when a parent is missing.
            parent = self.modules[parent_path]
            pkg_module = self.import_module(package, package, parent)
            curdir = pkg_module.__file__
            dirlist = os.listdir(curdir)
            for filename in dirlist:
                full = os.path.join(curdir, filename)
                if os.path.isdir(full):
                    if filename == "tests":
                        continue
                    init = os.path.join(full, "__init__.py")
                    if os.path.isfile(init):
                        stack.append((".".join((package, filename)), package))
                    continue
                if not filename.endswith(".py"):
                    continue
                if filename == "setup.py":  # skip
                    continue
                # We only accept .py files, so could use [:-3] too - faster...
                partname = os.path.splitext(filename)[0]
                self.import_module(partname, ".".join((package, partname)), pkg_module)

    def get_result(self):
        """Return 2-tuple: (list of packages, list of modules)."""
        keys = sorted(self.modules.keys())
        mods = []
        packs = []
        for key in keys:
            m = self.modules[key]
            if not m.__file__:  # skip builtins
                continue
            if m.__path__:
                packs.append(key)
            elif key != "__main__":
                mods.append(key)
        return (packs, mods)


if __name__ == "__main__":
    package = sys.argv[1]

    mf = CustomModuleFinder()
    mf.run_package(package)

    packs, mods = mf.get_result()

    print("Packages:")
    print(packs)

    print("Modules:")
    print(mods)
